/*
 * sid/kernel/sidkl26x86.c
 *
 * Copyright (C) 2004, 2005 Harald Deppeler
 *
 * hijacking sys_open for linux 2.6 kernels (on x86)
 * purpose: replace the receive_buf function for opened ptys so that
 *          we can hash/log keystrokes.
 *
 *  - most of this probably works for only for x86 - maybe we should catch that
 *  when compiling.
 *  - we rely on the presence of crypto modules in the kernel, specifically sha1
 *    (maybe we should catch that when loading the module)
 *
 *
 *  theory of operation
 *  -------------------
 *
 *  1) we adapt the address stored in position 5 in sys_call_table 
 *     (apparently sys_open) to the address of our wrapper function
 *     'sys_open_wrapper'
 *  2) sys_open_wrapper does our stuff and calls the original version of
 *     sys_open.
 *  3) upon unloading the module, we restore the original address (of
 *     sys_open) to position 5 in sys_call_table.
 *
 *  - as for grabbing pty input: this seems to differ from 2.4 kernels (see [2]).
 *    when oopsing in our new receive_buf function, we get a call stack like below.
 *   
 *    Dec 11 19:39:59 localhost kernel: Call Trace:
 *    Dec 11 19:39:59 localhost kernel:  [__alloc_pages+671/683] __alloc_pages+0x29f/0x2ab
 *    Dec 11 19:39:59 localhost kernel:  [__wake_up+33/68] __wake_up+0x21/0x44
 *    Dec 11 19:39:59 localhost kernel:  [pty_write+98/260] pty_write+0x62/0x104
 *    Dec 11 19:39:59 localhost kernel:  [write_chan+358/452] write_chan+0x166/0x1c4
 *    Dec 11 19:39:59 localhost kernel:  [default_wake_function+0/18] default_wake_function+0x0/0x12
 *    Dec 11 19:39:59 localhost kernel:  [default_wake_function+0/18] default_wake_function+0x0/0x12
 *    Dec 11 19:39:59 localhost kernel:  [tty_write+362/540] tty_write+0x16a/0x21c
 *    Dec 11 19:39:59 localhost kernel:  [write_chan+0/452] write_chan+0x0/0x1c4
 *    Dec 11 19:39:59 localhost kernel:  [vfs_write+175/217] vfs_write+0xaf/0xd9
 *    Dec 11 19:39:59 localhost kernel:  [sys_write+59/99] sys_write+0x3b/0x63
 *    Dec 11 19:39:59 localhost kernel:  [syscall_call+7/11] syscall_call+0x7/0xb
 *
 *    or:
 *
 *    Dec 12 18:31:16 localhost kernel:  [normal_poll+260/278] normal_poll+0x104/0x116
 *    Dec 12 18:31:16 localhost kernel:  [__alloc_pages+139/683] __alloc_pages+0x8b/0x2ab
 *    Dec 12 18:31:16 localhost kernel:  [copy_from_user+46/84] copy_from_user+0x2e/0x54
 *    Dec 12 18:31:16 localhost kernel:  [pty_write+160/260] pty_write+0xa0/0x104
 *    Dec 12 18:31:16 localhost kernel:  [write_chan+358/452] write_chan+0x166/0x1c4
 *    Dec 12 18:31:16 localhost kernel:  [default_wake_function+0/18] default_wake_function+0x0/0x12
 *    Dec 12 18:31:16 localhost kernel:  [default_wake_function+0/18] default_wake_function+0x0/0x12
 *    Dec 12 18:31:16 localhost kernel:  [tty_write+362/540] tty_write+0x16a/0x21c
 *    Dec 12 18:31:16 localhost kernel:  [write_chan+0/452] write_chan+0x0/0x1c4
 *    Dec 12 18:31:16 localhost kernel:  [vfs_write+175/217] vfs_write+0xaf/0xd9
 *    Dec 12 18:31:16 localhost kernel:  [sys_write+59/99] sys_write+0x3b/0x63
 *    Dec 12 18:31:16 localhost kernel:  [syscall_call+7/11] syscall_call+0x7/0xb
 *
 *
 *  - as for the ldisc functions: when examining the addresses for tty->ldisc.(...) for pty type ttys
 *    and looking it up in the system.map, we get the following:
 *    tty->ldisc.receive_buf: 	n_tty_receive_room 	(???)
 *    tty->ldisc.close: 	n_tty_close
 *    tty->ldisc.receive_room: 	n_tty_write_wakeup
 *    tty->ldisc.write: 	write_chan
 *    tty->ldisc.read: 		read_chan
 *    
 *
 *  References:
 *    [1] http://www.l0t3k.org/biblio/kernel/english/kernel-hijack.txt
 *    [2] http://www.phrack.org/phrack/59/p59-0x0e.txt
 */
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/stat.h>
#include <linux/err.h>
#include <linux/file.h>
#include <linux/fs.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/smp_lock.h>
#include <linux/mm.h>
#include <asm/scatterlist.h>
#include <linux/crypto.h>

unsigned long t_sys_open = 0;
unsigned long t_sys_call_table = 0;

#define MAX_BUFFER 81
#define MAX_SPECIAL_CHAR_SZ 12

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Harald Deppeler");

const char *mod_name = "sidkl26x86";

module_param(t_sys_call_table, ulong, S_IRUSR);
MODULE_PARM_DESC(t_sys_call_table, "address of sys_call_table");

struct sid {
	struct tty_struct *tty;
	char buf[MAX_BUFFER + MAX_SPECIAL_CHAR_SZ];
	int lastpos;
	int status;
	int pass;
	long lines;
};

/* FIXME: adapt this to a large enough number, like the hard limit for the
 * number of PTYs, if that exists */
#define MAXPTYS 200

struct sid *ttys[MAXPTYS] = { NULL };

/* function prototypes */
void (*old_n_tty_close) (struct tty_struct *);

void (*old_receive_buf) (struct tty_struct *, const unsigned char *,
			 char *, int);

static inline void init_tty(struct tty_struct *, int);

/* original sys_open function pointer */
long (*sys_open_fp) (const char *, int, int);

/* hash_and_log() - calculate sha1 hash of the line and log it 
 * @buf:
 * @size:
 * @tty:
 * @lines: */
int hash_and_log(char *buf, int size, int tty, long lines)
{
	int ret = 0;
	char *p;

	struct scatterlist sg[1];
	unsigned char digest[20];
	struct crypto_tfm *tfm;

	tfm = crypto_alloc_tfm("sha1", 0);
	if (tfm == NULL) {
		/* TODO */
		//fail();
		printk("%s: cryptoinit failed\n", mod_name);
		return -1;
	}

	p = buf;
	sg[0].page = virt_to_page(p);
	sg[0].offset = offset_in_page(p);
	sg[0].length = size;

	crypto_digest_init(tfm);
	crypto_digest_update(tfm, sg, 1);
	crypto_digest_final(tfm, digest);

	printk("tsw: tty=%ld line=%ld digest=%02x%02x%02x%02x "
	       "%02x%02x%02x%02x %02x%02x%02x%02x "
	       "%02x%02x%02x%02x %02x%02x%02x%02x\n",
	       (long)tty, lines, digest[0], digest[1],
	       digest[2], digest[3], digest[4], digest[5],
	       digest[6], digest[7], digest[8], digest[9],
	       digest[10], digest[11], digest[12], digest[13],
	       digest[14], digest[15], digest[16], digest[17],
	       digest[18], digest[19]);

	crypto_free_tfm(tfm);

	return ret;
}

/* sid_process_line() - fix up collected line and call hash_and_log() 
 * @tty:
 * @tmp:
 * @cont: */
void sid_process_line(struct tty_struct *tty, struct sid *tmp, int cont)
{
	char loginfo[MAX_BUFFER + MAX_SPECIAL_CHAR_SZ + 256];

	tmp->buf[tmp->lastpos++] = 0x0A;	/* newline */
	tmp->buf[tmp->lastpos] = 0;	/* == '\0' ? */

	/* tmp->status is 0 the first time this is called 
	 * why is this treated differently? */
	if (!tmp->status) {
		snprintf(loginfo, sizeof(loginfo) - 1, "%s", tmp->buf);
		(void)hash_and_log(loginfo, strlen(loginfo), tty->index,
				   tmp->lines);
	} else
		(void)hash_and_log(tmp->buf, tmp->lastpos, tty->index,
				   tmp->lines);

	tmp->lines++;

	tmp->status = 1;
}

/* sid_char() - collect chars and call line processing when necessary 
 * @tty:
 * @cp:
 * @count: */
void sid_char(struct tty_struct *tty, const unsigned char *cp, int count)
{
	struct sid *tmp = ttys[tty->index];

	/*printk("tsw: sid_char: tty %d character '%c'\n",
	   tty->index, *cp);*/

	if (!tmp) {
		//DPRINT("erm .. unknow error???\n");
		init_tty(tty, tty->index);
		tmp = ttys[tty->index];
		if (!tmp)
			return;
	}

	/* FIXME: process characters using the count argument */
	if ((count + tmp->lastpos) > MAX_BUFFER - 1 || *cp == '\r') {
		sid_process_line(tty, tmp, 1);
		tmp->buf[0] = '\0';
		tmp->lastpos = 0;
	} else {
		tmp->lastpos++;
		strncat(tmp->buf, cp, 1);
	}
}

/* new_n_tty_close(:) - log closing line and call original function
 * @tty:
 *
 * Description: - log a characteristic 'close line' for the user part to recognize 
 * 		  that this terminal has closed
 * 		- call the original close function */
void new_n_tty_close(struct tty_struct *tty)
{
	struct sid *tmp = ttys[tty->index];

	printk("tsw: tty=%d line=%ld digest=00000000 00000000 00000000 "
	       "00000000 00000000\n", tty->index, tmp->lines);
	(*old_n_tty_close) (tty);
}

/* new_receive_buf() - log the character passed and call original function 
 * @tty:
 * @cp:
 * @fp:
 * @count: 
 *
 * Note: when tty->read_buf == NULL, log only every other time this function
 * is called */
void new_receive_buf(struct tty_struct *tty, const unsigned char *cp,
		     char *fp, int count)
{
	static int flag[MAXPTYS];	/* FIXME: really no need to lock? */

	if (!tty->raw && count >= 0 && cp != NULL) {
		if (!tty->read_buf) {
			/*printk("new_receive_buf for tty->index %d count: %d,"
			       "flag: %d cp (at %08lx): \"%c\" "
			       "(!tty->read_buf)\n", tty->index, count,
		   	       flag[tty->index], (unsigned long)cp, *cp);*/ 
			if (flag[tty->index]) {
				sid_char(tty, cp, count);
				--flag[tty->index];
			} else
				++flag[tty->index];
		} else {
			/*printk("new_receive_buf for tty->index %d count: %d,"
			       "flag: %d cp (at %08lx): \"%c\" "
			       "(tty->read_buf)\n", tty->index, count,
		   	       flag[tty->index], (unsigned long)cp, *cp);*/
			sid_char(tty, cp, count);
		}
	}
	(*old_receive_buf) (tty, cp, fp, count);
}

/* init_tty() - replaces original receive_buf/tty_close with our own versions
 * @tty: tty to mangle
 * @tty_index: pty number
 * 
 * Description: for existing and new ttys
 * 		- set the tty receive_buf function to our own. 
 * 		- set the tty close function to our own 
 * 		- log a characteristic 'open line' for the user part to recognize
 * 		  that this terminal has opened */
static inline void init_tty(struct tty_struct *tty, int tty_index)
{
	struct sid *tmp;

	if (ttys[tty_index] == NULL) {	/* first time we open tty number tty_index */
		tmp = kmalloc(sizeof(struct sid), GFP_KERNEL);
		if (!tmp) {
			//DEBUG1(("kmalloc failed!\n"));
			return;
		}
		memset(tmp, 0, sizeof(struct sid));
		tmp->lines = 1;
		tmp->lastpos = 0;
		tmp->buf[0] = '\0';
		tmp->tty = tty;
		tty->ldisc.receive_buf = new_receive_buf;
		tty->ldisc.close = new_n_tty_close;
		ttys[tty_index] = tmp;
		printk("init_tty for tty->index %d\n", tty_index);
	} else {
		tmp = ttys[tty_index];
		tmp->lines = 1;
		tmp->buf[0] = '\0';
		tmp->lastpos = 0;
		tty->ldisc.receive_buf = new_receive_buf;
		tty->ldisc.close = new_n_tty_close;
		printk("init_tty for tty->index %d\n", tty_index);
	}
	printk("tsw: tty=%d line=0 digest=00000000 00000000 00000000 "
	       "00000000 00000000\n", tty->index);
}

/* sys_open_wrapper() - sys_open plus our stuff for pty opens 
 * @filename: 
 * @flags:
 * @mode:
 *
 * description: 1) call original sys_open
 *		2) when opening a pty, 
 *		   a) save function pointers for receive_buf 
 *		      and tty_close (exactly once)
 *                 b) initialize our data structure for the pty (once per pty)
 */
asmlinkage long sys_open_wrapper(const char *filename, int flags, int mode)
{
	long ret;
	struct tty_struct *tty;
	struct file *file;
	static int already = 0;

	ret = sys_open_fp(filename, flags, mode);

	if (ret >= 0) {
		/* FIXME: locking? */
		/*lock_kernel(); */
		file = fget(ret);
		tty = (struct tty_struct *)file->private_data;

		if (tty != NULL &&
		    tty->driver != NULL &&
		    tty->index >= 0 &&
		    tty->driver->type == TTY_DRIVER_TYPE_PTY &&
		    tty->driver->subtype == PTY_TYPE_SLAVE &&
		    tty->ldisc.receive_buf != NULL &&
		    tty->ldisc.receive_buf != new_receive_buf &&
		    tty->ldisc.close != NULL &&
		    tty->ldisc.close != new_n_tty_close) {
			spin_lock(already);
			if (!already) {
				old_receive_buf = tty->ldisc.receive_buf;
				old_n_tty_close = tty->ldisc.close;
				already = 1;
				printk("%s: receive_buf and n_tty_close are "
				       "at %08lx and %08lx, "
				       "respectively.\n",
				       mod_name,
				       (unsigned long)old_receive_buf,
				       (unsigned long)old_n_tty_close);
			}
			spin_unlock(already);
			init_tty(tty, tty->index);
		}

		/*file_list_lock(); */
		/*file_list_unlock(); */
		fput(file);
		/*unlock_kernel(); */
	}

	return ret;
}

static int __init sidkl26x86_init(void)
{
	long *p;
	printk(KERN_ALERT "loading %s ...\n", mod_name);

	if (0 == t_sys_call_table) {
        	printk(KERN_ALERT "%s: error: no sys_call_table provided. Try "
				"\"insmod %s.ko t_sys_call_table=0x...\" "
				"(... = awk '/sys_call_table/ { print $1 }'"
				"/boot/System.map-`uname -r`)\n", 
				mod_name, mod_name);
		return -1;
	}

	p = (long *)t_sys_call_table;
	p += 5;
	t_sys_open = *p;

	sys_open_fp = (long (*)(const char *, int, int))t_sys_open;

	printk(KERN_ALERT
	       "%s: modifiying sys_call_table at %08lx (sys_open): %08lx -> "
	       "%08lx\n", mod_name, (unsigned long)p, (unsigned long)*p,
	       (unsigned long)sys_open_wrapper);
	*p = (long)sys_open_wrapper;

	return 0;
}

static void __exit sidkl26x86_exit(void)
{
	long *p;
	int i;

	p = (long *)t_sys_call_table;
	p += 5;
	printk(KERN_ALERT
	       "%s: restoring sys_call_table at %08lx (sys_open): %08lx -> "
	       "%08lx\n", mod_name, (unsigned long)p, (unsigned long)*p,
	       t_sys_open);
	*p = (long)t_sys_open;

	for (i = 0; i < MAXPTYS; i++) {
		if (ttys[i] != NULL) {
			printk(KERN_ALERT
			       "%s: restoring receive_buf and tty_close for "
			       "tty %d\n", mod_name, i);
			ttys[i]->tty->ldisc.receive_buf = old_receive_buf;
			ttys[i]->tty->ldisc.close = old_n_tty_close;
			kfree(ttys[i]);
		}
	}
	printk(KERN_ALERT "... %s unloaded.\n", mod_name);
}

module_init(sidkl26x86_init);
module_exit(sidkl26x86_exit);
/* indent -kr -i8 -ts8 -sob -l80 -ss -ncs */
